//"use strict"

//----------------------------------
// imports
//----------------------------------

/**
 * /usr/share/gjs-1.0/
 * /usr/share/gnome-js/
 */
const Cairo = imports.cairo
const Lang = imports.lang
// http://developer.gnome.org/glib/unstable/glib-The-Main-Event-Loop.html
const Mainloop = imports.mainloop

/**
 * /usr/share/gjs-1.0/overrides/
 * /usr/share/gir-1.0/
 * /usr/lib/cinnamon/
 */
const Gio = imports.gi.Gio
const Gtk = imports.gi.Gtk
const Json = imports.gi.Json
const GLib = imports.gi.GLib
// http://developer.gnome.org/libsoup/stable/libsoup-client-howto.html
const Soup = imports.gi.Soup
// http://developer.gnome.org/st/stable/
const St = imports.gi.St

/**
 * /usr/share/cinnamon/js/
 */
const Applet = imports.ui.applet
const Config = imports.misc.config
const PopupMenu = imports.ui.popupMenu
const Settings = imports.ui.settings
const Util = imports.misc.util
const MessageTray = imports.ui.messageTray
const Main = imports.ui.main
const Cinnamon = imports.gi.Cinnamon

//----------------------------------------------------------------------
//
// Constants
//
//----------------------------------------------------------------------

const UUID = "cdos-patches@iscas.as.cn"

//----------------------------------------------------------------------
//
// Soup
//
//----------------------------------------------------------------------

// Soup session (see https://bugzilla.gnome.org/show_bug.cgi?id=661323#c64)
const _httpSession = new Soup.SessionAsync()
Soup.Session.prototype.add_feature.call(_httpSession, new Soup.ProxyResolverDefault())


//----------------------------------------------------------------------
//
// Logging
//
//----------------------------------------------------------------------

function log(message) {
    global.log(UUID + "#" + log.caller.name + ": " + message)
}

function logError(error) {
    global.logError(UUID + "#" + logError.caller.name + ": " + error)
}

//----------------------------------------------------------------------
//
// MyApplet
//
//----------------------------------------------------------------------

function MyApplet(orientation) {
    this._init(orientation)
}

MyApplet.prototype = {
    __proto__: Applet.IconApplet.prototype,

    _init: function (orientation) {
        Applet.IconApplet.prototype._init.call(this, orientation)

        try {
            this._source = null;
            this._update();
        }
        catch (e) {
            logError(e)
        }
    },
    on_applet_clicked: function(event) {
        let _appSys = Cinnamon.AppSystem.get_default();
        let _gsmApp = _appSys.lookup_app('cdospatchmgr.desktop');
        _gsmApp.activate();
    },
    _update: function() {
        //获取系统版本
        let [res, out, err, status] = GLib.spawn_command_line_sync('lsb_release -r')
        let sysversion = out.toString().substr(9, 3)
        let names = []
        let jsonobj
        let jsonobj1
        let contexthis = this

        //下载服务器补丁列表
        //this.loadJsonAsync('http://124.16.141.172:8080/cdospatches_server', function (json) {
        this.loadJsonAsync('http://127.0.0.1:3000/cospatches', function (json) {
            try {
                jsonobj = JSON.parse(json)
                //读文件，判断是否有更新。
                let f = Gio.file_new_for_path("/usr/share/cdospatchmgr/json.db")
                let loop = GLib.MainLoop.new(null, false);
                f.load_contents_async(null, function (f, res) {
                    let contents
                    try {
                        contents = f.load_contents_finish(res)[1]
                        jsonobj1 = JSON.parse(contents)
                        for (let i = 0, size = jsonobj.length; i < size; i++) {
                            if (jsonobj[i]['version'] == sysversion) {
                                for (let j = 0, size = jsonobj1.length; j < size; j++) {
                                    if (jsonobj[i]['location'] == jsonobj1[j]['location']) {
                                        break
                                    } else {
                                        names.push(jsonobj[i]['name'])
                                        break
                                    }
                                }
                            }
                        }
                        contexthis.set_applet_tooltip(_("system has")+ names.length +_("new pathces"))
                        if (names.length > 0){
                            contexthis.setMessage(_("system has")+ names.length +_("new pathces"))
                        }
                    } catch (e) {
                        log("*** ERROR: " + e.message)
                        loop.quit()
                        return
                    }
                    loop.quit()

                })
            }
            catch (e) {
                logError(e)
            }
        })


        Mainloop.timeout_add(2000000, Lang.bind(this, this._update))
       // this.set_applet_icon_path("/usr/share/cdospatchmgr/cdos-patches.png");
	this.set_applet_icon_symbolic_name("help-browser");
    },
    setMessage: function (text) {

    if(this._source == null) {
        this._source = new PATCHNotificationSource();
        this._source.connect('destroy', Lang.bind(this,
            function() {
                this._source = null;
            }));
        if (Main.messageTray) Main.messageTray.add(this._source);
    }
    let notification = null;
    if (this._source.notifications.length == 0) {
        notification = new MessageTray.Notification(this._source, text, null);
    } else {
        notification = this._source.notifications[0];
        notification.update(text, null, { clear: true });
    }
    notification.setTransient(true);
    this._source.notify(notification);
},
    loadJsonAsync: function loadJsonAsync(url, callback) {
        let context = this
        let message = Soup.Message.new('GET', url)
        _httpSession.queue_message(message, function soupQueue(session, message) {
            callback.call(context, message.response_body.data)
        })
    }
}

function PATCHNotificationSource() {
    this._init();
}

PATCHNotificationSource.prototype = {
    __proto__:  MessageTray.Source.prototype,

    _init: function() {
        MessageTray.Source.prototype._init.call(this, _("System Information"));

        this._setSummaryIcon(this.createNotificationIcon());
    },

    createNotificationIcon: function() {
        return new St.Icon({ icon_name: 'dialog-error',
            icon_type: St.IconType.FULLCOLOR,
            icon_size: this.ICON_SIZE });
    },

    open: function() {
        this.destroy();
    }
};

function main(metadata, orientation) {
    let myApplet = new MyApplet(orientation)
    return myApplet
}
