const St = imports.gi.St;
const ModalDialog = imports.ui.modalDialog;
const Gio = imports.gi.Gio;
const Gtk = imports.gi.Gtk;
const Lang = imports.lang;
const Clutter = imports.gi.Clutter;
const Applet = imports.ui.applet;
const PopupMenu = imports.ui.popupMenu;
const Cinnamon = imports.gi.Cinnamon;
const GTop = imports.gi.GTop;
const Mainloop = imports.mainloop;
const MessageTray = imports.ui.messageTray;
const Main = imports.ui.main;

function MyApplet(orientation, panel_height) {
    this._init(orientation, panel_height);
}

MyApplet.prototype = {
    __proto__: Applet.TextIconApplet.prototype,

    _init: function(orientation, panel_height) {        
        Applet.TextIconApplet.prototype._init.call(this, orientation, panel_height);
        
        try { 
	    this.gtop_cpu = new GTop.glibtop_cpu();
	    this.current = 0;
            this.last = 0;
            this.usage_cpu = 0;
            this.last_total = 0;
	    this.exceed_times = 0;

	    this._source = null;

            this._update();
        }
        catch (e) {
            global.logError(e);
        }
    },

    on_applet_clicked: function(event) {
           this._runSysMon();
    },

    _runSysMonActivate: function() {
           this._runSysMon();
    },

    _runSysMon: function() {
            let _appSys = Cinnamon.AppSystem.get_default();
            let _gsmApp = _appSys.lookup_app('gnome-system-monitor.desktop');
            _gsmApp.activate();
    },
    _update: function() {
            try {
	            GTop.glibtop_get_cpu(this.gtop_cpu);
                    this.current = this.gtop_cpu.idle;
                    let delta = (this.gtop_cpu.total - this.last_total) / 100;
                    if (delta > 0) {
                        this.usage_cpu = Math.round((this.current - this.last) / delta);
                        this.last = this.current;
                        this.last_total = this.gtop_cpu.total;
                    }

                    let percent = Math.round(100 - this.usage_cpu);
		    this.set_applet_tooltip(_("CPU:")+(percent.toString()).slice(-3) + "%");
                    if (percent > 95){
			this.exceed_times += 1;
			if (this.exceed_times >= 3){
	                    this.setMessage(_("Your CPU usage has exceeded 95%, the system performance will drop!"));
			    this.exceed_times = 0;
			}
                    }
                    Mainloop.timeout_add(2000, Lang.bind(this, this._update));
                }
            catch(e) {
                global.logError(e);
            }
	this.set_applet_icon_symbolic_name("utilities-system-monitor");
    },

    setMessage: function (text) {
	if(this._source == null) {
	    this._source = new CPUNotificationSource();
            this._source.connect('destroy', Lang.bind(this,
                function() {
                    this._source = null;
                }));
	    if (Main.messageTray) Main.messageTray.add(this._source);
	}
	let notification = null;
        if (this._source.notifications.length == 0) {
            notification = new MessageTray.Notification(this._source, text, null);
        } else {
            notification = this._source.notifications[0];
            notification.update(text, null, { clear: true });
        }
	notification.setTransient(true);
	this._source.notify(notification);
    },

    on_orientation_changed: function (orientation) {
        this._initContextMenu();
    }
};

function CPUNotificationSource() {
    this._init();
}

CPUNotificationSource.prototype = {
    __proto__:  MessageTray.Source.prototype,

    _init: function() {
        MessageTray.Source.prototype._init.call(this, _("System Information"));

        this._setSummaryIcon(this.createNotificationIcon());
    },

    createNotificationIcon: function() {
        return new St.Icon({ icon_name: 'dialog-error',
                             icon_type: St.IconType.FULLCOLOR,
                             icon_size: this.ICON_SIZE });
    },

    open: function() {
        this.destroy();
    }
};

function main(metadata, orientation, panel_height) {      
    let myApplet = new MyApplet(orientation, panel_height);
    return myApplet;      
}
